package de.unisaarland.cs.st.tautoko.completion;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.objectweb.asm.Type;

import de.unisaarland.cs.st.adabu.trans.model.MethodInvocation;
import de.unisaarland.cs.st.adabu.util.asm.MethodIdentifier;
import de.unisaarland.cs.st.adabu.util.asm.MyType;


public class BasicModifierGenerator {

	private List<String> importedClasses;

	private Map<String, Map<MethodIdentifier, List<InstrumentBlock>>> blocks;

	private String modifierClassName;

	private String modifierPackageName;

	private Set<Integer> instructionsToDelete;

	private static final String DEFAULT_PACKAGE = "de.unisaarland.cs.st.tautoko.instrument";

	public BasicModifierGenerator(String modifierClassName) {
		this(modifierClassName, DEFAULT_PACKAGE);
	}

	public BasicModifierGenerator(String modifierClassName, String modifierPackageName) {
		this.modifierClassName = modifierClassName;
		this.modifierPackageName = modifierPackageName;
		this.importedClasses = new Vector<String>();
		this.blocks = new HashMap<String, Map<MethodIdentifier, List<InstrumentBlock>>>();
		this.instructionsToDelete = new HashSet<Integer>();
		addDefaultImports();
	}

	private void addDefaultImports() {
		importedClasses.add(DEFAULT_PACKAGE + ".*");
		importedClasses.add("org.objectweb.asm.ClassWriter");
		importedClasses.add("org.objectweb.asm.tree.*");
		importedClasses.add("de.unisaarland.cs.st.adabu.util.asm.ASMTreeUtilities");
	}

	public Set<String> getInstrumentedClassNames() {
		return Collections.unmodifiableSet(blocks.keySet());
	}
	
	public void addBlock(MethodIdentifier methodToInstrument, String[] instructions, int insertPosition) {
		String className = methodToInstrument.getClassName();
		Map<MethodIdentifier, List<InstrumentBlock>> classBlocks = blocks.get(className);
		if (classBlocks == null) {
			classBlocks = new HashMap<MethodIdentifier, List<InstrumentBlock>>();
			blocks.put(className, classBlocks);
		}
		List<InstrumentBlock> methodBlocks = classBlocks.get(methodToInstrument);
		if (methodBlocks == null) {
			methodBlocks = new Vector<InstrumentBlock>();
			classBlocks.put(methodToInstrument, methodBlocks);
		}
		methodBlocks.add(new InstrumentBlock(instructions, insertPosition));
	}

	public void addInstructionToDelete(int instructionIndex) {
		instructionsToDelete.add(instructionIndex);
	}

	private void appendImports(StringBuffer code) {
		for (String importClass : importedClasses) {
			code.append("import " + importClass + ";\n");
		}
		code.append("\n");
	}

	public JavaClass generateClass() {
		StringBuffer code = new StringBuffer();
		code.append("package " + modifierPackageName + ";\n\n");
		appendImports(code);
		code.append("public class " + modifierClassName + " extends AbstractModifier {\n");

		//start transform method
		code.append("\t@Override\n" +
		"\tpublic byte[] transform(String className, byte[] bytes) {\n");
		for (String className : blocks.keySet()) {
			Map<MethodIdentifier, List<InstrumentBlock>> classBlocks = blocks.get(className);
			for (MethodIdentifier methodToInstrument: classBlocks.keySet()) {
				List<InstrumentBlock> blocks = classBlocks.get(methodToInstrument);
				//				start if block
				code.append("\t\tif (className.equals(\"" + methodToInstrument.getClassName() + "\")) {\n");
				code.append("\t\t\tClassNode classNode = ASMTreeUtilities.getClassNode(bytes);\n");
				code.append("\t\t\tMethodNode method = ASMTreeUtilities.getMethod(classNode, \"" + methodToInstrument.getMethodName() + "\", \"" + methodToInstrument.getSignature() + "\");\n");
				code.append("\t\t\tInsnList instructions = method.instructions;\n");
				//append position instructions for all blocks
				for (int blockIndex = 0; blockIndex < blocks.size(); blockIndex++) {
					InstrumentBlock block = blocks.get(blockIndex);
					code.append("\t\t\tAbstractInsnNode insertPosition" + blockIndex + " = instructions.get(" + block.getInsertPosition() + ");\n");
				}
				if (instructionsToDelete.size() > 0) {
					int deleteCounter = 0;
					for (Integer deleteIndex: instructionsToDelete) {
						code.append("\t\t\tAbstractInsnNode deletePosition" + deleteCounter + " = instructions.get(" + deleteIndex + ");\n");
						deleteCounter++;
					}
					for (int i = 0; i < deleteCounter; i++) {
						code.append("\t\t\tinstructions.remove(deletePosition" + i + ");\n");
					}
				}
				for (int blockIndex = 0; blockIndex < blocks.size(); blockIndex++) {
					InstrumentBlock block = blocks.get(blockIndex);
					String[] instructions = block.getInstructions();
					code.append("\t\t\t{\n");
					code.append("\t\t\t\tInsnList insertInstructions = new InsnList();\n");
					for (String instruction : instructions) {
						code.append("\t\t\t\t" + instruction + "\n");
					}
					code.append("\t\t\t\tinstructions.insert(insertPosition" + blockIndex + ", insertInstructions);\n");
					code.append("\t\t\t}\n");
				}
				//code.append("\t\t\tClassWriter writer = new ClassWriter(ClassWriter.COMPUTE_FRAMES + ClassWriter.COMPUTE_MAXS);\n");
				code.append("\t\t\tClassWriter writer = new ClassWriter(ClassWriter.COMPUTE_MAXS);\n");
				code.append("\t\t\tclassNode.accept(writer);\n");
				code.append("\t\t\treturn writer.toByteArray();\n");
				//end if block
				code.append("\t\t}\n");
			}
		}
		code.append("\t\treturn null;\n");
		//end transform method
		code.append("\t}\n");
		//end class
		code.append("}\n");
		return new JavaClass(modifierPackageName, modifierClassName, code.toString());
	}

	protected String getClassNameFromType(MyType type) {
		assert (type.isComplexType());
		return type.getTypeString().substring(1, type.getTypeString().length() - 1);
	}

	public void addSupressMethodCallBlock(MethodIdentifier methodToInstrument, MethodInvocation invocation, int methodCallInstructionIndex, String comment) {
		MethodIdentifier identifier = invocation.getIdentifier();
		int parameterStackHeight = getParameterStackHeight(identifier.getSignature());
		if (!identifier.isStatic()) {
			parameterStackHeight += 1;
		}
		List<String> instructions = new Vector<String>();
		if (comment != null) {
			instructions.add("//" + comment);
		}
		for (int i = 0; i < parameterStackHeight; i++) {
			instructions.add("insertInstructions.add(new InsnNode(POP));");
		}
		instructions.add(getLoadDefaultValueInstruction(identifier));
		addBlock(methodToInstrument, instructions.toArray(new String[instructions.size()]), methodCallInstructionIndex - 1);
		addInstructionToDelete(methodCallInstructionIndex);
	}

	public static int getParameterStackHeight(String desc) {
		Type[] argTypes = Type.getArgumentTypes(desc);
		int effect = 0;
		for (Type argType : argTypes) {
			effect += argType.getSize();
		}
		return effect;
	}

	public void addMethodCallBlock(MethodIdentifier methodToInstrument, MethodIdentifier methodToCall, MethodIdentifier precedingMethodCall, 
	                               int loadInstructionIndex, int methodCallInstructionIndex, String comment) {
		List<String> instructions = new Vector<String>();
		instructions.add("insertInstructions.add(new InsnNode(DUP));");
		addBlock(methodToInstrument, instructions.toArray(new String[instructions.size()]), loadInstructionIndex);
		instructions.clear();
		if (comment != null) {
			instructions.add("//" + comment);
		}
		if (precedingMethodCall.isFunction()) {
			instructions.add("insertInstructions.add(new InsnNode(SWAP));");
		}
		instructions.add("insertInstructions.add(new TypeInsnNode(CHECKCAST, \"" + methodToCall.getClassName() + "\"));");
		instructions.add("insertInstructions.add(new MethodInsnNode(INVOKEVIRTUAL, \"" + methodToCall.getClassName() + "\" , \"" + 
		                 methodToCall.getMethodName() + "\" , \"" + methodToCall.getSignature() + "\"));");
		if (methodToCall.isFunction()) {
			instructions.add("insertInstructions.add(new InsnNode(POP));");
		}
		addBlock(methodToInstrument, instructions.toArray(new String[instructions.size()]), methodCallInstructionIndex);
	}

	public List<String> getReturnDefaultValueInstructions(MethodIdentifier identifier) {
		List<String> instructions = new Vector<String>();
		String returnType = identifier.getReturnType();
		if (returnType.equals("V")) {
			instructions.add("insertInstructions.add(new InsnNode(RETURN));");
		} else if (returnType.startsWith("L") || returnType.startsWith("[")) {
			instructions.add("insertInstructions.add(new InsnNode(ACONST_NULL));");
			instructions.add("insertInstructions.add(new InsnNode(ARETURN));");
		} else if (returnType.equals("I") || returnType.equals("B") || returnType.equals("S") || returnType.equals("C") || returnType.equals("Z")) {
			instructions.add("insertInstructions.add(new InsnNode(ICONST_0));");
			instructions.add("insertInstructions.add(new InsnNode(IRETURN));");
		} else if (returnType.equals("J")) {
			instructions.add("insertInstructions.add(new InsnNode(LCONST_0));");
			instructions.add("insertInstructions.add(new InsnNode(LRETURN));");
		} else if (returnType.equals("D")) {
			instructions.add("insertInstructions.add(new InsnNode(DCONST_0));");
			instructions.add("insertInstructions.add(new InsnNode(DRETURN));");
		} else if (returnType.equals("F")) {
			instructions.add("insertInstructions.add(new InsnNode(FCONST_0));");
			instructions.add("insertInstructions.add(new InsnNode(FRETURN));");
		} else {
			throw new UnsupportedOperationException("Unhandled return type " + returnType);
		}
		return instructions;
	}

	public String getLoadDefaultValueInstruction(MethodIdentifier identifier) {
		String returnType = identifier.getReturnType();
		if (returnType.equals("V")) {
			return "insertInstructions.add(new InsnNode(NOP));";
		} else if (returnType.startsWith("L") || returnType.startsWith("[")) {
			return "insertInstructions.add(new InsnNode(ACONST_NULL));";
		} else if (returnType.equals("I") || returnType.equals("B") || returnType.equals("S") || returnType.equals("C") || returnType.equals("Z")) {
			return "insertInstructions.add(new InsnNode(ICONST_0));";
		} else if (returnType.equals("J")) {
			return "insertInstructions.add(new InsnNode(LCONST_0));";
		} else if (returnType.equals("D")) {
			return "insertInstructions.add(new InsnNode(DCONST_0));";
		} else if (returnType.equals("F")) {
			return "insertInstructions.add(new InsnNode(FCONST_0));";
		} else {
			throw new UnsupportedOperationException("Unhandled return type " + returnType);
		}
	}

	public Set<MethodIdentifier> getInstrumentedMethods() {
		Set<MethodIdentifier> result = new HashSet<MethodIdentifier>();
		for (String className : blocks.keySet()) {
			result.addAll(blocks.get(className).keySet());
		}
		return result;
	}

	public boolean hasBlocks() {
		return blocks.size() > 0;
	}

	public String getModifierClassName() {
		return modifierClassName;
	}

	public void setModifierClassName(String modifierClassName) {
		this.modifierClassName = modifierClassName;
	}
	
}
