package de.unisaarland.cs.st.tautoko.completion;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Random;
import java.util.Set;
import java.util.Vector;

import org.apache.commons.exec.ExecuteException;
import org.apache.commons.io.FileUtils;
import org.apache.log4j.Logger;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.Type;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.util.TraceClassVisitor;

import de.unisaarland.cs.st.adabu.core.CallStack;
import de.unisaarland.cs.st.adabu.core.deep.miner.TypeSetDeepTransitiveModelMiner;
import de.unisaarland.cs.st.adabu.core.heappath.HeapPath;
import de.unisaarland.cs.st.adabu.trans.abstraction.MissingFieldsAdder;
import de.unisaarland.cs.st.adabu.trans.abstraction.NeverChangingFieldsRemover;
import de.unisaarland.cs.st.adabu.trans.model.MethodInvocation;
import de.unisaarland.cs.st.adabu.trans.model.MethodInvocationSet;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectModel;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectState;
import de.unisaarland.cs.st.adabu.util.Pair;
import de.unisaarland.cs.st.adabu.util.asm.ASMTreeUtilities;
import de.unisaarland.cs.st.adabu.util.asm.ASMUtilities;
import de.unisaarland.cs.st.adabu.util.asm.MethodIdentifier;
import de.unisaarland.cs.st.adabu.util.datastructures.Graph;
import de.unisaarland.cs.st.adabu.util.datastructures.GraphEdge;
import de.unisaarland.cs.st.adabu.util.datastructures.GraphNode;
import de.unisaarland.cs.st.adabu.util.io.ObjectReaderWriter;
import de.unisaarland.cs.st.adabu.util.jung.EdgeContainer;
import de.unisaarland.cs.st.adabu.util.jung.IJungASSPAlgorithm;
import de.unisaarland.cs.st.adabu.util.jung.IJungDistanceMeasurer;
import de.unisaarland.cs.st.adabu.util.jung.JungASSPResult;
import de.unisaarland.cs.st.adabu.util.jung.JungFloydWarshall;
import de.unisaarland.cs.st.adabu.util.jung.JungPath;
import de.unisaarland.cs.st.adabu.util.jung.JungUtil;
import de.unisaarland.cs.st.adabu.util.source.ClassFile;
import de.unisaarland.cs.st.adabu.util.source.ClassFileSource;
import de.unisaarland.cs.st.adabu.util.source.ClassPathClassFileSource;
import de.unisaarland.cs.st.ibugs.util.Util;
import de.unisaarland.cs.st.ibugs.util.Util.ExecutionResult;
import de.unisaarland.cs.st.tautoko.ModelToTypeStateConverter;
import de.unisaarland.cs.st.tautoko.TypeStateGraphMLSerializer;
import de.unisaarland.cs.st.tautoko.abstraction.TypeStateStateAbstracter;

public class ModelCompleter {

	public static final int TEST_TIMEOUT = 240000;

	public static final String INSERT_MODIFIERCLASS = "Inserter";

	private HashSet<String> acceptedTypes;

	private String appClassPath;

	private String compileClassPath;

	private String testCommand;

	private File workingDirectory;

	private File tmpDirectory;

	private int heapPathLength;

	private int timeout;

	private static final Logger logger = Logger.getLogger(ModelCompleter.class);

	private Random random = new Random();

	private boolean useClassNamesForComplexFields = false;

	private Map<String, Set<TransitiveObjectModel>> typeToInitialModelMap;

	public ModelCompleter(File workingDirectory, File tmpDirectory, String testCommand, 
	                      String appClassPath, String compileClassPath, String acceptedType, 
	                      int heapPathLength, int timeout, boolean useClassNamesForComplexFields ) {
		this.workingDirectory = workingDirectory;
		this.tmpDirectory = tmpDirectory;
		this.testCommand = testCommand;
		this.appClassPath = appClassPath;
		this.compileClassPath = compileClassPath;
		this.acceptedTypes = new HashSet<String>();
		this.acceptedTypes.add(acceptedType);
		this.heapPathLength = heapPathLength;
		this.timeout = timeout;
		this.useClassNamesForComplexFields = useClassNamesForComplexFields;
	}

	public Map<String, Set<TransitiveObjectModel>> process() throws ExecuteException, IOException {
		ClassFileSource source = new ClassPathClassFileSource(appClassPath);
		File traceFile = new File(tmpDirectory + "/trace.ser");
		ExecutionResult executionResult = traceTests(testCommand, traceFile, workingDirectory, timeout);
		if (isTestOk(executionResult)) {
			List<TransitiveObjectModel> initialModels = mineModels(traceFile);
			abstractModels(initialModels, source, useClassNamesForComplexFields);
			typeToInitialModelMap = getTypeToModelMap(initialModels);
			Map<String, Set<TransitiveObjectModel>> completedModels = completeModels(initialModels, source);
			return completedModels;
		} else {
			logger.error("Unable to successfully run test suite to generate initial models. Process output follows.");
			logger.error(executionResult.getOut());
			return null;
		}
	}
	
 public static Set<String> getSuperTypes(String type, ClassFileSource source) throws IOException {
		Set<String> result = new HashSet<String>();
		result.add(type);
		do {
			try {
				ClassFile classFile = source.getClassForName(type);
				type = ASMUtilities.getSuperClassName(classFile.getClassBytes());
				if (type != null) {
					result.add(type);
				}
			} catch (NoSuchElementException ex) {
				logger.warn("Unable to resolve class " + type);
				return result;
			}
		} while (type != null);
		return result;
	}
	
	public static Set<MethodIdentifier> getAllPublicNonStaticMethods(String className, ClassFileSource source) throws NoSuchElementException, IOException {
		Set<String> superTypes = getSuperTypes(className, source);
		superTypes.remove("java/lang/Object");
		Set<MethodIdentifier> methods = new HashSet<MethodIdentifier>();
		for (String type : superTypes) {
			methods.addAll(ASMUtilities.getAllMethods(source.getClassForName(type).getClassBytes()));
		}
		Set<MethodIdentifier> result = new HashSet<MethodIdentifier>();
		for (MethodIdentifier methodIdentifier : methods) {
			if (methodIdentifier.isPublic() && !methodIdentifier.isStatic() && !methodIdentifier.isConstructor() && 
					!(methodIdentifier.getMethodName().equals("clone") && methodIdentifier.getSignature().startsWith("()")) &&
					!(methodIdentifier.getMethodName().equals("toString") && methodIdentifier.getSignature().equals("()Ljava/lang/String;"))) {
				result.add(methodIdentifier);
			}
		}
		return result;
	}

	public Map<String, Set<TransitiveObjectModel>> completeModels(List<TransitiveObjectModel> models, ClassFileSource source) throws NoSuchElementException, IOException {
		Set<JavaClass> javaClasses  = new HashSet<JavaClass>();
		Map<MethodIdentifier, Set<TransitiveObjectState>> methodToInvokedStatesMap = getMethodToInvokedStatesMap(models);
		for (TransitiveObjectModel model: models) {
			logger.info("Investigating model " + model);
			Set<MethodIdentifier> allCallableMethods = getAllPublicNonStaticMethods(model.getClassName(), source);
			Set<TransitiveObjectState> nonStartingStates = getAllNonStartingStates(model);
			//Set<MethodInvocation> allMethodInvocations = getInvokedMethods(model);
			//Map<MethodIdentifier, Set<MethodInvocation>> identifierToInvocationsMap = getIdentifierToInvocationsMap(allMethodInvocations);
			for (MethodIdentifier currentMethod : allCallableMethods) {
				if (!currentMethod.isConstructor()) {
					Set<TransitiveObjectState> statesToInvestigate = new HashSet<TransitiveObjectState>(nonStartingStates);
					Set<TransitiveObjectState> originatingStates = getAllOriginatingStates(model, currentMethod);
					Set<TransitiveObjectState> exercisedStates = methodToInvokedStatesMap.get(currentMethod);
					if (exercisedStates == null) {
						exercisedStates = new HashSet<TransitiveObjectState>();
						methodToInvokedStatesMap.put(currentMethod, exercisedStates);
					}
					statesToInvestigate.removeAll(originatingStates);
					statesToInvestigate.removeAll(exercisedStates);
					if (!statesToInvestigate.isEmpty()) {
						logger.debug("Found " + statesToInvestigate.size() + " states to investigate for method " + currentMethod);
						for (TransitiveObjectState stateToInvestigate : statesToInvestigate) {
							try {
								if (!currentMethod.hasParameters()) {
									List<Pair<BasicModifierGenerator, MethodInvocation>> insertMethodCallPairs = insertMethodCall(model, currentMethod, stateToInvestigate, source);
									if (insertMethodCallPairs.size() > 0) {
										exercisedStates.add(stateToInvestigate);
										for (Pair<BasicModifierGenerator, MethodInvocation> pair : insertMethodCallPairs) {
											javaClasses.add(pair.getU().generateClass());
										}
									}
								} else {
									List<BasicModifierGenerator> supressMethodCallGenerators = supressMethodCalls(model, currentMethod, stateToInvestigate, source, originatingStates);
									for (BasicModifierGenerator basicModifierGenerator : supressMethodCallGenerators) {
										javaClasses.add(basicModifierGenerator.generateClass());
									}
								}
							} catch (IOException exception) {
								logger.warn("Exception occured processing state to investigate. This state will be ignored.", exception);
							}
						}
					} else {
						logger.debug("All states exercised for method " + currentMethod);
					}
				}
			}
		}
		List<JavaClass> sortedJavaClasses = new Vector<JavaClass>(javaClasses);
		Collections.sort(sortedJavaClasses, new Comparator<JavaClass>() {

			public int compare(JavaClass o1, JavaClass o2) {
				return o1.getContent().length() - o2.getContent().length();
			}

		});
		return processJavaClasses(sortedJavaClasses, models, source);
	}

	public static List<BasicModifierGenerator> supressMethodCalls(TransitiveObjectModel model, MethodIdentifier currentMethod,
	                                                              TransitiveObjectState stateToInvestigate, ClassFileSource source,
	                                                              Set<TransitiveObjectState> originatingStates) throws IOException {
		List<BasicModifierGenerator> result = new Vector<BasicModifierGenerator>();
		edu.uci.ics.jung.graph.DirectedGraph<TransitiveObjectState, EdgeContainer<TransitiveObjectState, MethodInvocationSet>> jungGraph = JungUtil.convertToJungSparseMultiGraph(model.getModel());
		IJungASSPAlgorithm<TransitiveObjectState, EdgeContainer<TransitiveObjectState, MethodInvocationSet>> alg = new JungFloydWarshall<TransitiveObjectState, EdgeContainer<TransitiveObjectState,MethodInvocationSet>>();
		JungASSPResult<TransitiveObjectState, EdgeContainer<TransitiveObjectState, MethodInvocationSet>> allShortestPaths = alg.getShortestPaths(jungGraph, new IJungDistanceMeasurer<TransitiveObjectState, EdgeContainer<TransitiveObjectState,MethodInvocationSet>>() {
			public double getEdgeLength(TransitiveObjectState startNode,
			                            TransitiveObjectState endNode,
			                            edu.uci.ics.jung.graph.Graph<TransitiveObjectState, EdgeContainer<TransitiveObjectState, MethodInvocationSet>> graph) {
				return 1d;
			}
		});				
		for (TransitiveObjectState transitiveObjectState : originatingStates) {
			List<JungPath<TransitiveObjectState, EdgeContainer<TransitiveObjectState, MethodInvocationSet>>> shortestPaths = allShortestPaths.getAllShortestPaths(stateToInvestigate, transitiveObjectState);
			if (shortestPaths != null) {
				for (JungPath<TransitiveObjectState, EdgeContainer<TransitiveObjectState, MethodInvocationSet>> jungPath : shortestPaths) {
					List<MethodInvocation> invocationsToSupress = new Vector<MethodInvocation>();
					for (EdgeContainer<TransitiveObjectState, MethodInvocationSet> edgeContainer : jungPath) {
						invocationsToSupress.add(getBestMethodToSuppress(edgeContainer.getData()));
					}
					boolean containsConstructorInvocation = false;
					for (MethodInvocation methodInvocation : invocationsToSupress) {
						if (methodInvocation.getIdentifier().isConstructor()) {
							containsConstructorInvocation = true;
							break;
						}
					}
					if (!containsConstructorInvocation) {
						BasicModifierGenerator generator = new BasicModifierGenerator(INSERT_MODIFIERCLASS);
						for (MethodInvocation methodInvocation : invocationsToSupress) {
							CallStack callStack = methodInvocation.getCallStack();
							if (callStack.size() > 1) {
								MethodIdentifier callingMethod = callStack.getCallStack().get(callStack.size() - 2).getIdentifier();
								InsnList instructions = getMethodInstructions(source, callingMethod);
								//displayClass(callingMethod.getClassName(), source);
								try {
									int methodCallInstructionIndex = getMethodCallPosition(methodInvocation.getCallSiteId(), instructions, methodInvocation.getIdentifier());
									generator.addSupressMethodCallBlock(callingMethod, methodInvocation, methodCallInstructionIndex, "Suppressing invocation of method " + methodInvocation.getIdentifier());
								} catch (IOException insertException) {
									logger.error("An error occured determining method call position. This may have to do with methods that cannot be instrumented. Please check.", insertException);
								}
							} else {
								logger.warn("Unable to find invoking method for stack " + callStack);
							}
						}
						if (generator.hasBlocks()) {
							result.add(generator);
						}
					}
				}
			}
		}
		return result;
	}

	private static void displayClass(String className, ClassFileSource source) throws NoSuchElementException, IOException {
		ClassFile classFile = source.getClassForName(className);
		displayClass(classFile.getClassBytes());
	}

	private static void displayClass(byte[] classBytes) {
		ClassReader classReader;
		TraceClassVisitor traceVisitor;

		classReader = new ClassReader(classBytes);
		traceVisitor = new TraceClassVisitor(new PrintWriter(System.out));
		classReader.accept(traceVisitor, ClassReader.EXPAND_FRAMES);
	}   



	public static MethodInvocation getBestMethodToSuppress(MethodInvocationSet invocationSet) {
		List<MethodInvocation> invocations = new Vector<MethodInvocation>(invocationSet.getInvocations());
		Collections.sort(invocations, new Comparator<MethodInvocation>(){
			public int compare(MethodInvocation o1, MethodInvocation o2) {
				MethodIdentifier firstIdentifier = o1.getIdentifier();
				MethodIdentifier secondIdentifier = o2.getIdentifier();
				boolean isFirstVoid = firstIdentifier.getReturnType().equals("V");
				boolean isSecondVoid = secondIdentifier.getReturnType().equals("V");
				if (isFirstVoid ^ isSecondVoid ) {
					if (isFirstVoid) {
						return -1;
					} else {
						return +1;
					}
				} else {
					Type[] firstParameters = Type.getArgumentTypes(firstIdentifier.getSignature());
					Type[] secondParameters = Type.getArgumentTypes(secondIdentifier.getSignature());
					return firstParameters.length - secondParameters.length;
				}
			}
		});
		return invocations.get(0);
	}


	protected static void saveModel(Graph<String, List<MethodIdentifier>> typeState, File resultFile) throws IOException {
		ObjectReaderWriter<Graph<String, List<MethodIdentifier>>> writer = new ObjectReaderWriter<Graph<String,List<MethodIdentifier>>>();
		writer.write(typeState, resultFile);
	}

	protected static void saveModel(TransitiveObjectModel model, File resultFile) throws IOException {
		ObjectReaderWriter<TransitiveObjectModel> writer = new ObjectReaderWriter<TransitiveObjectModel>();
		writer.write(model, resultFile);
	}	

	protected Map<String, Set<TransitiveObjectModel>> processJavaClasses(List<JavaClass> javaClasses, 
	                                                                     List<TransitiveObjectModel> models,
	                                                                     ClassFileSource source) throws IOException {
		logger.info("Checking " + javaClasses.size() + " modifications.");
		Set<TransitiveObjectModel>  allModels = new HashSet<TransitiveObjectModel>();
		for (TransitiveObjectModel transitiveObjectModel : models) {
			allModels.add(transitiveObjectModel);
		}
		int javaClassCounter = 1;
		for (JavaClass javaClass : javaClasses) {
			FileUtils.cleanDirectory(tmpDirectory);
			//if (javaClassCounter != 20) {
			//javaClassCounter++;
			//logger.warn("Skipping modification number " + javaClassCounter);
			//continue;
			//}
			logger.trace("Checking modification number " + javaClassCounter + " out of " + javaClasses.size() + " :");
			logger.trace(javaClass.toString());
			String modifierClassName = javaClass.getPackageName() + "." + javaClass.getClassName();
			ExecutionResult result = compile(javaClass);

			if (result.getResult() == 0) {
				File traceFile = new File(tmpDirectory + "/trace.ser");
				result = traceModifiedTests(testCommand, modifierClassName, traceFile, workingDirectory, tmpDirectory, compileClassPath);
				logger.info("Result of tracing the modified test: " + result.getResult());
				logger.debug(result.getOut());
				try {
					List<TransitiveObjectModel> newModels = mineModels(traceFile);
					abstractModels(newModels, source, useClassNamesForComplexFields);
					int newModelCounter = 0;
					for (TransitiveObjectModel transitiveObjectModel : newModels) {
						if (!allModels.contains(transitiveObjectModel)) {
							logger.info("Found new model: ");
							logger.info(transitiveObjectModel.getModel().createDottyRepresentation());
							newModelCounter++;
							allModels.add(transitiveObjectModel);
						}
					}
					logger.info("Got " + newModelCounter + " new models from test.");
				} catch (Exception exception) {
					logger.error("An exception occured processing trace for java class " + javaClassCounter + ". This trace will be ignored.", exception);
				}
			} else {
				logger.error("Unable to compile class:");
				logger.error(result.getOut());
			}
			javaClassCounter++;
			//logger.info("Warning:Skipping after one modification.");
			//break;
		}
		Map<String, Set<TransitiveObjectModel>> typeToModelMap = getTypeToModelMap(allModels);
		return typeToModelMap;
	}

	public static Map<String, Set<TransitiveObjectModel>> getTypeToModelMap(Collection<TransitiveObjectModel> allModels) {
		Map<String, Set<TransitiveObjectModel>> typeToModelMap = new HashMap<String, Set<TransitiveObjectModel>>();
		for (TransitiveObjectModel transitiveObjectModel : allModels) {
			String type = transitiveObjectModel.getClassName();
			Set<TransitiveObjectModel> typeModels = typeToModelMap.get(type);
			if (typeModels == null) {
				typeModels = new HashSet<TransitiveObjectModel>();
				typeToModelMap.put(type, typeModels);
			}
			typeModels.add(transitiveObjectModel);
		}
		return typeToModelMap;
	}



	public static List<Pair<BasicModifierGenerator, MethodInvocation>> insertMethodCall(TransitiveObjectModel model, MethodIdentifier methodToCall, 
	                                                                                    TransitiveObjectState stateToInvestigate, ClassFileSource source) throws NoSuchElementException, IOException {
		List<Pair<BasicModifierGenerator, MethodInvocation>> result = new Vector<Pair<BasicModifierGenerator,MethodInvocation>>();
		GraphNode<TransitiveObjectState, MethodInvocationSet> nodeToInvestigate = model.getModel().getNodeForData(stateToInvestigate);
		Vector<GraphNode<TransitiveObjectState, MethodInvocationSet>> predecessorNodes = nodeToInvestigate.getPredecessorNodes();
		for (int i = 0; i < predecessorNodes.size(); i++) {
			GraphNode<TransitiveObjectState, MethodInvocationSet> predecessorNode = predecessorNodes.get(i);
			GraphEdge<TransitiveObjectState, MethodInvocationSet> edge = predecessorNode.getOutgoingEdgeToNode(nodeToInvestigate);
			for (MethodInvocation invocation: edge.getData()) {
				CallStack callStack = invocation.getCallStack();
				if (callStack.size() > 1) {
					MethodIdentifier precedingMethodCall = callStack.getActiveMethod().getIdentifier();
					MethodIdentifier callingMethod = callStack.getCallStack().get(callStack.size() -2).getIdentifier();
					int callSiteId = invocation.getCallSiteId();
					InsnList instructions = getMethodInstructions(source, callingMethod);
					//logger.trace("Instructions for method  " + callingMethod.getMethodName());
					//logger.trace(ASMTreeUtilities.getOpCodeString(instructions));
					//if (precedingMethodCall.getMethodName().equals("uidStore") && methodToCall.getMethodName().equals("noop")) {
						//System.out.println("Meep");
					//}
					int methodCallInstructionIndex = getMethodCallPosition(callSiteId, instructions, invocation.getIdentifier());
					if (methodCallInstructionIndex != -1) {
						int thisLoadInstructionIndex = getThisLoadInstructionIndex(instructions, methodCallInstructionIndex, invocation.getIdentifier());
						if (thisLoadInstructionIndex != -1) {
							BasicModifierGenerator generator = new BasicModifierGenerator(INSERT_MODIFIERCLASS);
							generator.addMethodCallBlock(callingMethod, methodToCall, precedingMethodCall, 
							                             thisLoadInstructionIndex, methodCallInstructionIndex,
							                             "call to " + methodToCall + " after " + precedingMethodCall);
							result.add(new Pair<BasicModifierGenerator, MethodInvocation>(generator, invocation));
						}
					}
				}
			}
		}
		return result;
	}

	public static InsnList getMethodInstructions(ClassFileSource source, MethodIdentifier method) throws IOException {
		ClassFile classFile = source.getClassForName(method.getClassName());
		ClassNode classNode = ASMTreeUtilities.getClassNode(classFile.getClassBytes());
		MethodNode methodNode = ASMTreeUtilities.getMethod(classNode, method.getMethodName(), method.getSignature());
		InsnList instructions = methodNode.instructions;
		return instructions;
	}

	public static int getParameterStackHeight(String desc) {
		Type[] argTypes = Type.getArgumentTypes(desc);
		int effect = 0;
		for (Type argType : argTypes) {
			effect += argType.getSize();
		}
		return effect;
	}

	public static int getThisLoadInstructionIndex(InsnList instructions, int methodCallIndex, MethodIdentifier calledMethodIdentifier) {
		// + 1 is needed because we know that those method calls are INVOKEVIRTUAL
		int parameterStackEffect = getParameterStackHeight(calledMethodIdentifier.getSignature()) + 1;
		int stackHeight = 0;
		for (int instructionIndex = methodCallIndex - 1; instructionIndex >= 0; instructionIndex--) {
			AbstractInsnNode instruction = instructions.get(instructionIndex);
			try {
				int effect = StackEffects.getEffectOnStack(instruction);
				stackHeight += effect;
				if (stackHeight == parameterStackEffect) {
					//it may be the case that there are subsequent instructions that load the reference we need,
					//for example, ALOAD,GETFIELD 
					//For those ugly cases, we need to adjust the instruction index
					int correctedInstructionIndex = instructionIndex;
					for (; correctedInstructionIndex < methodCallIndex - 1; correctedInstructionIndex++) {
						AbstractInsnNode correctedInstruction = instructions.get(correctedInstructionIndex + 1);
						int correctedEffect = StackEffects.getEffectOnStack(correctedInstruction);
						if (correctedEffect != 0) {
							return correctedInstructionIndex;
						}
					}
					//this special case is needed for sequence aload getfield invoke
					if (correctedInstructionIndex != instructionIndex) {
						return correctedInstructionIndex;
					} else {
						return instructionIndex;
					}
				}
			} catch (RuntimeException exception) {
				logger.debug("An error occured identifying this load instruction", exception);
			}
		}
		return -1;
	}

	protected List<Integer> getLineNumberPositions(MethodIdentifier identifier, ClassFileSource source) throws NoSuchElementException, IOException {
		ClassFile classFile = source.getClassForName(identifier.getClassName());
		ClassNode classNode = ASMTreeUtilities.getClassNode(classFile.getClassBytes());
		MethodNode methodNode = ASMTreeUtilities.getMethod(classNode, identifier.getMethodName(), identifier.getSignature());
		List<Integer> result = ASMTreeUtilities.getLineNumberPositions(methodNode);
		Collections.sort(result);
		return result;
	}

	public static boolean isInstrumentedCallSite(String className) {
		return !className.startsWith("java/lang/String") && !className.startsWith("java/lang/adabu/Tracer");
	}

	public static boolean isReflectionInvocation(String className, String methodName) {
		return (className.equals("java/lang/reflect/Method") && methodName.equals("invoke")) ||
		(className.equals("java/lang/reflect/Constructor") && methodName.equals("newInstance")) ||
		(className.equals("java/lang/Class") && methodName.equals("newInstance"));
	}

	public static Integer getMethodCallPosition(int callSiteId, InsnList instructions, MethodIdentifier calledMethod) throws NoSuchElementException, IOException {
		int callSiteCounter = -1;
		for (int i = 0; i < instructions.size(); i++) {
			AbstractInsnNode instruction = instructions.get(i);
			if (instruction instanceof MethodInsnNode) {
				MethodInsnNode methodCall = (MethodInsnNode) instruction;
				if (isInstrumentedCallSite(methodCall.owner)) {
					callSiteCounter++;
					if (callSiteCounter == callSiteId) {
						if (calledMethod != null) {
							if (!methodCall.name.equals(calledMethod.getMethodName()) || ! methodCall.desc.equals(calledMethod.getSignature())) {
								if (isReflectionInvocation(methodCall.owner, methodCall.name)) {
									logger.warn("Expected call to method " + calledMethod + " at position " + callSiteId + " but found reflection call " + methodCall.owner + "." + methodCall.name + methodCall.desc + " . Trying with reflection invocation.");
								} else {
									throw new IOException("Expected call to method " + calledMethod + " at position " + callSiteId + " but found call to " + methodCall.owner + "." + methodCall.name + methodCall.desc);
								}
							}
						}
						return i;
					}
				}
			}
		}
		return -1;
	}

	public static Set<TransitiveObjectState> getAllNonStartingStates(TransitiveObjectModel model) {
		Set<TransitiveObjectState> result = new HashSet<TransitiveObjectState>();
		for (GraphNode<TransitiveObjectState, MethodInvocationSet> node : model.getModel().getNodes()) {
			if (!node.getData().equals(model.getStartingState())) {
				result.add(node.getData());
			}
		}
		return result;
	}

	public static Set<TransitiveObjectState> getAllOriginatingStates(TransitiveObjectModel objectModel,
	                                                                 MethodIdentifier methodIdentifier) {
		Set<TransitiveObjectState> states = new HashSet<TransitiveObjectState>();
		Graph<TransitiveObjectState, MethodInvocationSet> model = objectModel.getModel(); 
		Vector<GraphEdge<TransitiveObjectState, MethodInvocationSet>> edges = model.getEdges();
		for (GraphEdge<TransitiveObjectState, MethodInvocationSet> graphEdge : edges) {
			if (graphEdge.getData().getAllInvocationsOfMethod(methodIdentifier).size() > 0) {
				states.add(graphEdge.getStartNode().getData());
			}
		}	
		return states;
	}

	public static Map<MethodIdentifier, Set<MethodInvocation>> getIdentifierToInvocationsMap(Set<MethodInvocation> allInvocations) {
		Map<MethodIdentifier, Set<MethodInvocation>> result = new HashMap<MethodIdentifier, Set<MethodInvocation>>();
		for (MethodInvocation methodInvocation : allInvocations) {
			MethodIdentifier identifier = methodInvocation.getIdentifier();
			Set<MethodInvocation> invocationsForMethod = result.get(identifier);
			if (invocationsForMethod == null) {
				invocationsForMethod = new HashSet<MethodInvocation>();
				result.put(identifier, invocationsForMethod);
			}
			invocationsForMethod.add(methodInvocation);
		}
		return result;
	}

	public static Set<MethodInvocation> getInvokedMethods(TransitiveObjectModel objectModel) {
		Set<MethodInvocation> invokedMethods = new HashSet<MethodInvocation>();
		Graph<TransitiveObjectState, MethodInvocationSet> model = objectModel.getModel(); 
		Vector<GraphEdge<TransitiveObjectState, MethodInvocationSet>> edges = model.getEdges();
		for (GraphEdge<TransitiveObjectState, MethodInvocationSet> graphEdge : edges) {
			Set<MethodInvocation> methodInvocations = graphEdge.getData().getInvocations();
			for (MethodInvocation methodInvocation : methodInvocations) {
				invokedMethods.add(methodInvocation);
			}
		}
		return invokedMethods;
	}

	public static void abstractModels(List<TransitiveObjectModel> models, ClassFileSource source, boolean useClassNamesForComplexFields) {
		for (int i = 0; i < models.size(); i++) {
			TransitiveObjectModel objectModel = models.get(i);
			MissingFieldsAdder adder = new MissingFieldsAdder(objectModel);
			objectModel = adder.createAbstractModel();
			//we are not allowed to remove fields that don't change at any given point
			//NeverChangingFieldsRemover remover = new NeverChangingFieldsRemover(objectModel);
			//objectModel = remover.createAbstractModel();
			TypeStateStateAbstracter abstracter = new TypeStateStateAbstracter(objectModel, source, useClassNamesForComplexFields);
			objectModel = abstracter.createAbstractModel();
			objectModel.setObjectId(-1);
			models.set(i, objectModel);			
		}	
	}

	public List<TransitiveObjectModel> mineModels(File traceFile) throws IOException {
		TypeSetDeepTransitiveModelMiner miner = new TypeSetDeepTransitiveModelMiner(traceFile, heapPathLength, null, acceptedTypes);
		miner.readTrace(traceFile);
		List<TransitiveObjectModel> models = miner.getModels();
		return models;
	}

	public static boolean isTestOk(ExecutionResult executionResult) {
		return executionResult.getResult() == 0;
	}

	public static Map<MethodIdentifier, Set<TransitiveObjectState>> getMethodToInvokedStatesMap(List<TransitiveObjectModel> models) {
		Map<MethodIdentifier, Set<TransitiveObjectState>> 	result = new HashMap<MethodIdentifier, Set<TransitiveObjectState>>();
		for (TransitiveObjectModel transitiveObjectModel : models) {
			for (GraphEdge<TransitiveObjectState, MethodInvocationSet> edge : transitiveObjectModel.getModel().getEdges()) {
				for (MethodInvocation invocation : edge.getData()) {
					MethodIdentifier identifier = invocation.getIdentifier();
					Set<TransitiveObjectState> methodStates = result.get(identifier);
					if (methodStates == null) {
						methodStates = new HashSet<TransitiveObjectState>();
						result.put(identifier, methodStates);
					}
					methodStates.add(edge.getStartNode().getData());
				}
			}
		}
		return result;
	}

	protected ExecutionResult compile(JavaClass javaClass) throws IOException {
		File packageDirectory = new File(tmpDirectory.getAbsolutePath() + File.separator + javaClass.getPackageName().replace('.', '/'));
		packageDirectory.mkdirs();
		File classFile = new File(packageDirectory.getAbsolutePath() + File.separator + javaClass.getClassName() + ".java");
		FileUtils.writeStringToFile(classFile, javaClass.getContent(), null);
		String compiler;
		if (System.getProperty("os.name").equalsIgnoreCase("linux")) {
			compiler  = "/usr/lib/jvm/jdk1.6.0/bin/javac";
		} else {
			compiler = "/scratch2/dallmeier/soylatte/bin/javac";	
		}
		List<String> args = new Vector<String>();
		args.add("-cp");
		args.add(compileClassPath + ":.");
		args.add(javaClass.getPackageName().replace('.', '/') + "/" + javaClass.getClassName() + ".java");
		return Util.runCommand(compiler, args.toArray(new String[args.size()]), tmpDirectory, logger);
	}

	public static ExecutionResult traceTests(String testCommand, File traceFile, File workingDirectory, int timeout) throws ExecuteException, IOException {
		Vector<String> args = new Vector<String>();
		args.add("-Dadabu.resultfilename=" + traceFile.getAbsolutePath());
		args.add("-Dadabu.debug=false");
		//args.add("-Dadabu.displaypattern=.*UserFileLoginModule");
		args.add("-Dadabu.exclusionpattern=.*ClassLoader.*|java/lang/.*|java/io/.*|sun/misc.*|sun/reflect/.*|sun/net/.*|sun/instrument/.*");
		args.add("-Xbootclasspath/a:/scratch2/dallmeier/bin/projects/adabu/adabu-tracer/target/adabu-tracer-0.5.jar:" +
		         "/scratch2/dallmeier/bin/projects/adabu/adabu-core/target/classes:" +
		         "/scratch2/dallmeier/bin/repository/commons-collections/commons-collections/3.2/commons-collections-3.2.jar:" +
		         "/scratch2/dallmeier/bin/repository/asm/asm-tree/3.2/asm-tree-3.2.jar:" +
		         "/scratch2/dallmeier/bin/repository/asm/asm/3.2/asm-3.2.jar:" + 
		         "/scratch2/dallmeier/bin/repository/asm/asm-commons/3.2/asm-commons-3.2.jar:" + 
		"/scratch2/dallmeier/bin/repository/asm/asm-util/3.2/asm-util-3.2.jar");
		args.add("-javaagent:/scratch2/dallmeier/bin/projects/adabu/adabu-tracer/target/adabu-tracer-0.5.jar");
		return Util.runCommand(testCommand, args.toArray(new String[args.size()]), workingDirectory, logger, timeout);
	}

	public ExecutionResult traceModifiedTests(String testCommand, String modifierClassName, 
	                                          File traceFile, File workingDirectory,
	                                          File tmpDirectory, String compileClassPath) throws ExecuteException, IOException {
		String killIdentifier = Long.toString(random.nextLong());
		logger.debug("Tracing modified test with kill identifier " + killIdentifier);
		Vector<String> args = new Vector<String>();
		args.add("-Dadabu.resultfilename=" + traceFile.getAbsolutePath());
		args.add("-Dadabu.debug=false");
		args.add("-Dkill.identifier=" + killIdentifier);
		args.add("-Dadabu.exclusionpattern=.*ClassLoader.*|java/lang/.*|java/io/.*|sun/misc.*|sun/reflect/.*|sun/net/.*|sun/instrument/.*|org/objectweb/asm/.*");
		//args.add("-Dadabu.exclusionpattern=de/unisaarland/.*|org/objectweb/asm/.*|.*ClassLoader.*|java/lang/.*|java/io/.*|sun.*");
		args.add("-Dadabu.modifier=" + modifierClassName);
		args.add("-Xbootclasspath/a:/scratch2/dallmeier/bin/projects/adabu/adabu-tracer/target/adabu-tracer-0.5.jar:" +
		         "/scratch2/dallmeier/bin/projects/adabu/adabu-core/target/classes:" +
		         "/scratch2/dallmeier/bin/repository/commons-collections/commons-collections/3.2/commons-collections-3.2.jar:" +
		         "/scratch2/dallmeier/bin/repository/asm/asm-tree/3.2/asm-tree-3.2.jar:" +
		         "/scratch2/dallmeier/bin/repository/asm/asm/3.2/asm-3.2.jar:" + 
		         "/scratch2/dallmeier/bin/repository/asm/asm-commons/3.2/asm-commons-3.2.jar:" + 
		         "/scratch2/dallmeier/bin/repository/asm/asm-util/3.2/asm-util-3.2.jar:" +
		         tmpDirectory.getAbsolutePath() + ":" + compileClassPath);
		args.add("-javaagent:/scratch2/dallmeier/bin/projects/typestate/tautoko-core/target/tautoko-core-0.1-SNAPSHOT.jar");
		args.add("-javaagent:/scratch2/dallmeier/bin/projects/adabu/adabu-tracer/target/adabu-tracer-0.5.jar");
		return Util.runCommand(testCommand, args.toArray(new String[args.size()]), workingDirectory, logger, timeout, killIdentifier);
	}

	public Map<String, Set<TransitiveObjectModel>> getTypeToInitialModelMap() {
		return typeToInitialModelMap;
	}
}
