package de.unisaarland.cs.st.tautoko.completion;

import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.IincInsnNode;
import org.objectweb.asm.tree.InsnNode;
import org.objectweb.asm.tree.IntInsnNode;
import org.objectweb.asm.tree.JumpInsnNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MultiANewArrayInsnNode;
import org.objectweb.asm.tree.TypeInsnNode;
import org.objectweb.asm.tree.VarInsnNode;

import de.unisaarland.cs.st.adabu.util.asm.ASMUtilities;

public class StackEffects implements Opcodes {

	public static int getEffectOnStack(AbstractInsnNode insn) {
		if (insn instanceof InsnNode) {
			return getEffectOnStack((InsnNode) insn);
		} else if (insn instanceof FieldInsnNode) {
			return getEffectOnStack((FieldInsnNode) insn);
		} else if (insn instanceof IincInsnNode) {
			return 0;
		} else if (insn instanceof JumpInsnNode) {
			return getEffectOnStack((JumpInsnNode) insn);
		} else if (insn instanceof IntInsnNode) {
			return getEffectOnStack((IntInsnNode) insn);
		} else if (insn instanceof LdcInsnNode) {
			return getEffectOnStack((LdcInsnNode) insn);
		} else if (insn instanceof MethodInsnNode) {
			return getEffectOnStack((MethodInsnNode) insn);
		} else if (insn instanceof MultiANewArrayInsnNode) {
			return getEffectOnStack((MultiANewArrayInsnNode) insn);
		} else if (insn instanceof TypeInsnNode) {
			return getEffectOnStack((TypeInsnNode) insn);
		} else if (insn instanceof VarInsnNode) {
			return getEffectOnStack((VarInsnNode) insn);
		} else {
			return 0;
		}
	}

	public static int getParameterStackEffect(String desc) {
		Type[] argTypes = Type.getArgumentTypes(desc);
		Type returnType = Type.getReturnType(desc);
		int effect = 0;
		for (Type argType : argTypes) {
			effect -= argType.getSize();
		}
		if (!(returnType.getDescriptor().equals("V"))) {
			effect += returnType.getSize();
		}
		return effect;
	}

	public static int getEffectOnStack(MultiANewArrayInsnNode node) {
		return -node.dims + 1;
	}
	
	public static int getEffectOnStack(VarInsnNode insn) {
		int effect;
		int opCode = insn.getOpcode();
		switch(opCode) {
		case ILOAD: {
			effect = 1;
			break;
		}
		case LLOAD: {
			effect = 2;
			break;
		}
		case FLOAD: {
			effect = 1;
			break;
		}
		case DLOAD: {
			effect = 2;
			break;
		}
		case ALOAD: {
			effect = 1;
			break;
		}
		case ISTORE: {
			effect = -1;
			break;
		}
		case LSTORE: {
			effect = -2;
			break;
		}
		case DSTORE: {
			effect = -2;
			break;
		}
		case FSTORE: {
			effect = -1;
			break;
		}
		case ASTORE: {
			effect = -1;
			break;
		}
		default: {
			throw new RuntimeException();
		}	
		}
		return effect;
	}


	public static int getEffectOnStack(TypeInsnNode insn) {
		int effect;
		int opCode = insn.getOpcode();
		switch(opCode) {
		case NEW: {
			effect = 1;
			break;
		}
		case ANEWARRAY: {
			effect = 0;
			break;
		}
		case CHECKCAST: {
			effect = 0;
			break;
		}
		default: {
			throw new RuntimeException();
		}	
		}
		return effect;
	}

	public static int getEffectOnStack(MethodInsnNode insn) {
		int effect;
		int opCode = insn.getOpcode();
		int parameterReturnEffect = getParameterStackEffect(insn.desc);
		switch(opCode) {
		case INVOKESTATIC: {  
			effect = parameterReturnEffect;  
			break;
		}
		case INVOKEVIRTUAL:
		case INVOKESPECIAL:
		case INVOKEINTERFACE: {
			effect = parameterReturnEffect - 1;
			break;
		}
		default: {
			throw new RuntimeException();
		}
		}
		return effect;
	}

	public static int getEffectOnStack(LdcInsnNode insn) {
		if (insn.cst instanceof Integer) {
			return 1;
		} else if (insn.cst instanceof Float) {
			return 1;
		} else if (insn.cst instanceof Long) {
			return 2;
		} else if (insn.cst instanceof Double) {
			return 2;
		} else if (insn.cst instanceof String) {
			return 1;
		} else {
			throw new RuntimeException();
		}
	}

	public static int getEffectOnStack(JumpInsnNode insn) {
		int effect;
		int opCode = insn.getOpcode();
		switch(opCode) {
		case IFEQ: {  
			effect = -1;  
			break;
		}
		case IFNE: {
			effect = -1;
			break;
		}
		case IFLT: {
			effect = -1;
			break;
		}
		case IFGE: {
			effect = -1;
			break;
		}
		case IFGT: {
			effect = -1;
			break;
		}
		case IFLE: {
			effect = -1;
			break;
		}
		case IF_ICMPEQ: {
			effect = -2;
			break;
		}
		case IF_ICMPNE: {
			effect = -2;
			break;
		}
		case IF_ICMPLT: {
			effect = -2;
			break;
		}
		case IF_ICMPGE: {
			effect = -2;
			break;
		}
		case IF_ICMPGT: {
			effect = -2;
			break;
		}
		case IF_ICMPLE: {
			effect = -2;
			break;
		}
		case IF_ACMPEQ: {
			effect = -2;
			break;
		}
		case IF_ACMPNE: {
			effect = -2;
			break;
		}
		case GOTO: {
			effect = 0;
			break;
		}
		case JSR: {
			effect = 0;
			break;
		}
		case IFNULL : { 
			effect = -1;
			break;
		}
		case IFNONNULL: {
			effect = -1;
			break;
		}

		default: {
			throw new RuntimeException();
		}
		}
		return effect;
	}

	public static int getEffectOnStack(IntInsnNode insn) {
		int effect;
		int opCode = insn.getOpcode();
		switch(opCode) {
		case BIPUSH: {
			effect = 1;
			break;
		}
		case NEWARRAY: {
			effect = 0;
			break;
		}
		case SIPUSH: {
			effect = 1;
			break;
		}
		default: {
			throw new RuntimeException();
		}
		}
		return effect;
	}

	public static int getEffectOnStack(FieldInsnNode insn) {
		int effect;
		int opCode = insn.getOpcode();
		int size = ASMUtilities.getTypeSize(insn.desc);
		switch(opCode) {
		case GETSTATIC: {
			effect = size;
			break;
		}
		case PUTFIELD: {
			effect = -(size + 1);
			break;
		}
		case GETFIELD: {
			effect = size - 1;
			break;
		}
		case PUTSTATIC: {
			effect = -size;
			break;
		}
		default: {
			throw new RuntimeException();
		}
		}
		return effect;
	}

	public static int getEffectOnStack(InsnNode insn) {
		int effect;
		int opCode = insn.getOpcode();
		switch(opCode) {
		case NOP: {
			effect = 0;
			break;
		}
		case ACONST_NULL: {
			effect = 1;
			break;
		}
		case ICONST_M1: {
			effect = 1;
			break;
		}
		case ICONST_0: {
			effect = 1;
			break;
		}
		case ICONST_1: {
			effect = 1;
			break;
		}
		case ICONST_2: {
			effect = 1;
			break;
		}
		case ICONST_3: {
			effect = 1;
			break;
		}
		case ICONST_4: {
			effect = 1;
			break;
		}
		case ICONST_5: {
			effect = 1;
			break;
		}
		case LCONST_0: {
			effect = 2;
			break;
		}
		case LCONST_1: {
			effect = 2;
			break;
		}
		case FCONST_0: {
			effect = 1;
			break;
		}
		case FCONST_1: {
			effect = 1;
			break;
		}
		case FCONST_2: {
			effect = 1;
			break;
		}
		case DCONST_0: {
			effect = 2;
			break;
		}
		case DCONST_1: {
			effect = 2;
			break;
		}
		case IALOAD: {
			effect = -1;
			break;
		}
		case LALOAD: {
			effect = 0;
			break;
		}
		case FALOAD: {
			effect = -1;
			break;
		}
		case DALOAD: {
			effect = 0;
			break;
		}
		case AALOAD: {
			effect = -1;
			break;
		}
		case BALOAD: {
			effect = -1;
			break;
		}
		case CALOAD: {
			effect = -1;
			break;
		}
		case SALOAD: {
			effect = -1;
			break;
		}
		case IASTORE: {
			effect = -3;
			break;
		}
		case LASTORE: {
			effect = -4;
			break;
		}
		case FASTORE: {
			effect = -3;
			break;
		}
		case DASTORE: {
			effect = -4;
			break;
		}
		case AASTORE: {
			effect = -3;
			break;
		}
		case BASTORE: {
			effect = -3;
			break;
		}
		case CASTORE: {
			effect = -3;
			break;
		}
		case SASTORE: {
			effect = -3;
			break;
		}
		case POP: {
			effect = -1;
			break;
		}
		case POP2: {
			effect = -2;
			break;
		}
		case DUP: {
			effect = 1;
			break;
		}
		case DUP_X1: {
			effect = 1;
			break;
		}
		case DUP_X2: {
			effect = 1;
			break;
		}
		case DUP2: {
			effect = 2;
			break;
		}
		case DUP2_X1: {
			effect = 2;
			break;
		}
		case DUP2_X2: {
			effect = 2;
			break;
		}
		case SWAP: {
			effect = 0;
			break;
		}
		case IADD: {
			effect = -1;
			break;
		}
		case LADD: {
			effect = -2;
			break;
		}
		case FADD: {
			effect = -1;
			break;
		}
		case DADD: {
			effect = -2;
			break;
		}
		case ISUB: {
			effect = -1;
			break;
		}
		case LSUB: {
			effect = -2;
			break;
		}
		case FSUB: {
			effect = -1;
			break;
		}
		case DSUB: {
			effect = -2;
			break;
		}
		case IMUL: {
			effect = -1;
			break;
		}
		case LMUL: {
			effect = -2;
			break;
		}
		case FMUL: {
			effect = -1;
			break;
		}
		case DMUL: {
			effect = -2;
			break;
		}
		case IDIV: {
			effect = -1;
			break;
		}
		case LDIV: {
			effect = -2;
			break;
		}
		case FDIV: {
			effect = -1;
			break;
		}
		case DDIV: {
			effect = -2;
			break;
		}
		case IREM: {
			effect = -1;
			break;
		}
		case LREM: {
			effect = -2;
			break;
		}
		case FREM: {
			effect = -1;
			break;
		}
		case DREM: {
			effect = -2;
			break;
		}
		case INEG: {
			effect = 0;
			break;
		}
		case LNEG: {
			effect = 0;
			break;
		}
		case FNEG: {
			effect = 0;
			break;
		}
		case DNEG: {
			effect = 0;
			break;
		}
		case ISHL: {
			effect = -1;
			break;
		}
		case LSHL: {
			effect = -1;
			break;
		}
		case ISHR: {
			effect = -1;
			break;
		}
		case LSHR: {
			effect = -1;
			break;
		}
		case IUSHR: {
			effect = -1;
			break;
		}
		case LUSHR: {
			effect = -1;
			break;
		}
		case IAND: {
			effect = -1;
			break;
		}
		case LAND: {
			effect = -2;
			break;
		}
		case IOR: {
			effect = -1;
			break;
		}
		case LOR: {
			effect = -2;
			break;
		}
		case IXOR: {
			effect = -1;
			break;
		}
		case LXOR: {
			effect = -2;
			break;
		}
		case I2L: {
			effect = 1;
			break;
		}
		case I2F: {
			effect = 0;
			break;
		}
		case I2D: {
			effect = 1;
			break;
		}
		case L2I: {
			effect = -1;
			break;
		}
		case L2F: {
			effect = -1;
			break;
		}
		case L2D: {
			effect = 0;
			break;
		}
		case F2I: {
			effect = 0;
			break;
		}
		case F2L: {
			effect = 1;
			break;
		}
		case F2D: {
			effect = 1;
			break;
		}
		case D2I: {
			effect = -1;
			break;
		}
		case D2L: {
			effect = 0;
			break;
		}
		case D2F: {
			effect = -1;
			break;
		}
		case I2B: {
			effect = 0;
			break;
		}
		case I2C: {
			effect = 0;
			break;
		}
		case I2S: {
			effect = 0;
			break;
		}
		case LCMP: {
			effect = -3;
			break;
		}
		case FCMPL: {
			effect = -1;
			break;
		}
		case FCMPG: {
			effect = -1;
			break;
		}
		case DCMPL: {
			effect = -3;
			break;
		}
		case DCMPG: {
			effect = -3;
			break;
		}
		case ARRAYLENGTH: {
			effect = 0;
			break;
		}
		case MONITORENTER: {
			effect = -1;
			break;
		}
		case MONITOREXIT: {
			effect = -1;
			break;
		}
		default: {
			throw new RuntimeException();
		}
		}
		return effect;
	}
}
	