package de.unisaarland.cs.st.tautoko.verify;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.io.FileUtils;

import de.unisaarland.cs.st.adabu.util.Pair;
import de.unisaarland.cs.st.adabu.util.asm.MethodIdentifier;
import de.unisaarland.cs.st.ibugs.util.Util;
import de.unisaarland.cs.st.ibugs.util.Util.ExecutionResult;

public class JFTADriver implements IVerifierDriver {

	private String testCommand;
	
	private String specFileNamePrefix = "spec";
	
	private String specFileNameSuffix = ".xml";

	private boolean warnOnUnknownTransition;

	public JFTADriver(String testCommand, boolean warnOnUnknownTransition) {
		this.testCommand = testCommand;
		this.warnOnUnknownTransition = warnOnUnknownTransition;
	}

	private File writeSpecificationsToTmpDirectory(Collection<String> specifications, File tmpDirectory) throws IOException {
		File specDirectory = new File(tmpDirectory + "/specs");
		specDirectory.mkdirs();
		FileUtils.cleanDirectory(specDirectory);
		List<File> result = new Vector<File>(specifications.size());
		int i = 0;
		for (String specification : specifications) {
			File specFile = new File(specDirectory + "/" + specFileNamePrefix + i + specFileNameSuffix);
			FileUtils.writeStringToFile(specFile, specification);
			result.add(specFile);
			i++;
		}
		return specDirectory;
	}
	
	private static Pattern violationPattern = Pattern.compile("\\s*(\\d*): \\[([^\\]]*)\\]([\\w/\\$]*)@(\\d*).*", Pattern.DOTALL);
	
	public Pair<ExecutionResult, List<Violation>> testVerifierOnMethods(Collection<MethodIdentifier> methods, String classPath, String spec, File tmpDirectory) throws IOException {
		File specFiles = writeSpecificationsToTmpDirectory(Collections.singletonList(spec), tmpDirectory);
		List<String> params = getDefaultParams(classPath, spec, tmpDirectory, specFiles);
		for (MethodIdentifier identifier: methods) {
			params.add("--testmethod ");
			params.add(identifier.getClassName() + "." + identifier.getMethodName() + identifier.getSignature());
		}
		String[] paramArray = params.toArray(new String[params.size()]);
		ExecutionResult result = Util.runCommand(testCommand, paramArray, tmpDirectory, null);
		if (result.getResult() != 0) {
			throw new IOException("Verifier execution returned error code " + result.getResult());
		}
		List<Violation> violations = parseViolationsFromOutput(result.getOut());
		return new Pair<ExecutionResult, List<Violation>>(result, violations);
		
	}
	
	private List<String> getDefaultParams(String classPath, String spec, File tmpDirectory, File specFiles) {
		List<String> params = new Vector<String>();
		StringTokenizer classPathSplitter = new StringTokenizer(classPath, ":");
		while (classPathSplitter.hasMoreTokens()) {
			String entry = classPathSplitter.nextToken();
			params.add("--classfiles");
			params.add(entry);
		}
		params.add("--specfiles");
		params.add(specFiles.getAbsolutePath());
		params.add("--maxDepth");
		params.add("0");
		params.add("--unknownTransition");
		if (warnOnUnknownTransition) {
			params.add("error");
		} else {
			params.add("ignore");
		}
		params.add("--loglevel");
		params.add("DEBUG");
		return params;
	}
	
	public Pair<ExecutionResult, List<Violation>> testVerifierOnClasses(Collection<String> classNames, String classPath, String spec, File tmpDirectory) throws IOException {
		File specFiles = writeSpecificationsToTmpDirectory(Collections.singletonList(spec), tmpDirectory);
		List<String> params = getDefaultParams(classPath, spec, tmpDirectory, specFiles);
		for (String className : classNames) {
			params.add("--testclass ");
			params.add(className);
		}
		String[] paramArray = params.toArray(new String[params.size()]);
		ExecutionResult result = Util.runCommand(testCommand, paramArray, tmpDirectory, null);
		if (result.getResult() != 0) {
			throw new IOException("Verifier execution returned error code " + result.getResult());
		}
		List<Violation> violations = parseViolationsFromOutput(result.getOut());
		return new Pair<ExecutionResult, List<Violation>>(result, violations);
	}
	
	private List<Violation> parseViolationsFromOutput(String output) {
		List<Violation> violations = new Vector<Violation>();
		String[] lines = output.split("\n");
		for (String line : lines) {
			Matcher matcher = violationPattern.matcher(line);
			if (matcher.matches()) {
				String level = matcher.group(2);
				String className = matcher.group(3);
				int lineNumber = Integer.parseInt(matcher.group(4));
				Violation violation = new Violation(className, lineNumber);
				violation.setLevel(level);
				violations.add(violation);
			}
		}
		return violations;
	}
	
	public static void main(String[] args) {
		String testString = "  0: [WARN]org/columba/ristretto/imap/IMAPProtocolTest@127, call stack [testCapabilitiy]:\n";
		//String testString = "  0: [WARN]org/columba/ristretto/imap/IMAPProtocol$Test@127, ";
		//Pattern pattern = Pattern.compile("\\s*(\\d*): \\[([^\\]]*)\\]([^@]*)@(\\d*),.*", Pattern.DOTALL);
		Pattern pattern = Pattern.compile("\\s*(\\d*): \\[([^\\]]*)\\]([\\w/\\$]*)@(\\d*).*", Pattern.DOTALL);
		Matcher matcher = pattern.matcher(testString);
		System.out.println(matcher.matches());
		if (matcher.matches()) {
			int groupId = 1;
			System.out.println("Group " + groupId + ": " + matcher.group(groupId));
			groupId = 2;
			System.out.println("Group " + groupId + ": " + matcher.group(groupId));
			groupId = 3;
			System.out.println("Group " + groupId + ": " + matcher.group(groupId));
			groupId = 4;
			System.out.println("Group " + groupId + ": " + matcher.group(groupId));
		}
	}
}
